/*
 * Copyright 2011-2021 Connect2id Ltd. and others
 * Copyright 2021 by DTV-Verkehrsconsult, Aachen
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 * Contact Information:
 * DTV-Verkehrsconsult GmbH
 * Pascalstraße 53
 * 52076 Aachen, Germany
 * phone: +49 2408 7047 0
 * mail: <info@dtv-verkehrsconsult.de>
 */
package de.bsvrz.sys.funclib.srp6;


import java.math.BigInteger;
import java.security.MessageDigest;


/**
 * Alternative routine for computing a password key x = H(s | H(I | ":" | P))
 * 
 * <p>Specification: RFC 5054.
 *
 * <p>This routine can be passed to the {@link SRP6VerifierGenerator} and
 * {@link SRP6ClientSession} to replace the
 * {@link SRP6Routines#computeX default routine} x = H(s | H(P)).
 *
 * @author Vladimir Dzhuvinov
 */
public class XRoutineWithUserIdentity implements XRoutine {


	/**
	 * Computes x = H(s | H(I | ":" | P))
	 *
	 * <p>This method complies with the RFC 5054 recommendation, save for 
	 * the hash algorithm which can be any (RFC 5054 recommends SHA-1).
	 *
	 * @param digest   The hash function 'H'. Must not be {@code null}.
	 * @param salt     The salt 's'. Must not be {@code null}.
	 * @param username The user identity 'I'. Must not be {@code null}.
	 * @param password The user password 'P'. Must not be {@code null}.
	 *
	 * @return The resulting 'x' value.
	 */
	public BigInteger computeX(final MessageDigest digest,
				   final byte[] salt,
				   final byte[] username,
				   final byte[] password) {
	
		digest.update(username);
		digest.update((byte)':');
		digest.update(password);
		
		byte[] output = digest.digest();
		
		digest.update(salt);
		output = digest.digest(output);

		return BigIntegerUtils.bigIntegerFromBytes(output);
	}
	
	
	/**
	 * Returns a string representation of this routine algorithm.
	 * 
	 * @return "H(s | H(I | ":" | P))"
	 */
	public String toString() {
	
		return "H(s | H(I | \":\" | P))";
	}
}
